use lazy_static::lazy_static;
use regex::Regex;
use std::{process::Command, str::FromStr};

const pm_cfg: &str = include_str!("../../package_manager.list");

/// regex: `(?<pm>[a-z]+)>(?<action>(?:un)?install+): (?<command>.*)`
/// example: pacman>install: pacman -Sy %args

lazy_static! {
    static ref PM_REGEX: Regex =
        Regex::new(r"(?P<pm>[a-z]+)>(?P<action>(?:un)?install+): (?P<command>.*)").unwrap();
}

#[derive(Debug)]
struct PackageManager {
    name: String,
    command: String,
    args: Vec<String>,
}

impl PackageManager {
    fn new(name: &str, command: &str, args: Vec<String>) -> Self {
        Self {
            name: name.to_string(),
            command: command.to_string(),
            args,
        }
    }

    pub fn install(&self, packages: Vec<String>) -> Result<(), Vec<String>> {
        Command::new(&self.command)
            .args(&self.args)
            .args(packages)
            .spawn()
            .expect("err");
        Ok(())
    }
    pub fn uninstall(&self, packages: Vec<String>) -> Result<(), Vec<String>> {
        todo!();
    }
}
impl FromStr for PackageManager {
    type Err = String;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let caps = PM_REGEX.captures(s).ok_or("invalid package manager")?;
        let name = caps.name("pm").ok_or("invalid package manager")?.as_str();
        let command = caps
            .name("command")
            .ok_or("invalid package manager")?
            .as_str();
        let args = caps
            .name("args")
            .ok_or("invalid package manager")?
            .as_str()
            .split_whitespace()
            .map(|s| s.to_string())
            .collect();
        Ok(Self::new(name, command, args))
    }
}
#[derive(Debug)]
struct Package {
    name: String,
}
impl FromStr for Package {
    type Err = String;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let caps = PM_REGEX.captures(s).ok_or("invalid package")?;
        let name = caps.name("name").ok_or("invalid package")?.as_str();
        Ok(Self::new(name))
    }
}

impl Package {
    fn new(name: &str) -> Self {
        Self {
            name: name.to_string(),
        }
    }
}

pub fn package_managers() -> Vec<PackageManager> {
    pm_cfg
        .lines()
        .map(|s| s.to_string())
        .map(|s| s.parse::<PackageManager>())
        .collect::<Result<Vec<PackageManager>, String>>()
        .expect("invalid package manager")
}
